from typing import Optional, Union
from llama_index.core.llms.llm import LLM
from langchain.base_language import BaseLanguageModel

from llama_kg.llm_predictor.KronLangChainLLM import KronLangChainLLM
from llama_index.llms.openai import OpenAI

from llama_index.core.llms.utils import LLMType


def kron_resolve_llm(llm: Optional[LLMType] = None) -> LLM:
    if isinstance(llm, BaseLanguageModel):
        # NOTE: if it's a langchain model, wrap it in a LangChainLLM
        return KronLangChainLLM(llm=llm)

    return llm or OpenAI()



from llama_index.core.bridge.langchain import BaseLanguageModel, BaseChatModel
from llama_index.llms.langchain import LangChainLLM
from llama_index.core.bridge.langchain import OpenAI, ChatOpenAI

from llama_index.core.llms import LLMMetadata

from llama_kg.llm_predictor.openai_utils import kron_openai_modelname_to_contextsize

def is_chat_model(llm: BaseLanguageModel) -> bool:
    return isinstance(llm, BaseChatModel)

class KronLangChainLLM(LangChainLLM):
    """Adapter for a LangChain LLM."""

    def __init__(self, llm: BaseLanguageModel) -> None:
        super().__init__(llm)


    @property
    def metadata(self) -> LLMMetadata:
        is_chat_model_ = is_chat_model(self.llm)
        if isinstance(self.llm, OpenAI):
            return LLMMetadata(
                context_window=kron_openai_modelname_to_contextsize(self.llm.model_name),
                num_output=self.llm.max_tokens,
                is_chat_model=is_chat_model_ ,
            )
        elif isinstance(self.llm, ChatOpenAI):
            return LLMMetadata(
                context_window=kron_openai_modelname_to_contextsize(self.llm.model_name),
                num_output=self.llm.max_tokens or -1,
                is_chat_model=is_chat_model_ ,
            )
        else:
            return super().metadata()




from typing import Any, Generator, Optional, Protocol, Tuple, runtime_checkable

from llama_index.core.service_context_elements.llm_predictor import  LLMPredictor
from llama_index.core.llms.utils import LLMType
from llama_index.core.callbacks.base import CallbackManager

from llama_kg.llm_predictor.utils import kron_resolve_llm

class KronLLMPredictor(LLMPredictor):
    """LLM predictor class.

    Wrapper around an LLMChain from Langchain.

    Args:
        llm (Optional[langchain.llms.base.LLM]): LLM from Langchain to use
            for predictions. Defaults to OpenAI's text-davinci-003 model.
            Please see `Langchain's LLM Page
            <https://langchain.readthedocs.io/en/latest/modules/llms.html>`_
            for more details.

        retry_on_throttling (bool): Whether to retry on rate limit errors.
            Defaults to true.

        cache (Optional[langchain.cache.BaseCache]) : use cached result for LLM
    """

    def __init__(
        self,
        llm: Optional[LLMType] = None,
        callback_manager: Optional[CallbackManager] = None,
     ) -> None:
        """Initialize params."""
        super().__init__(llm=llm, callback_manager=callback_manager)
        self._llm = kron_resolve_llm(llm)
        #self.callback_manager = callback_manager or CallbackManager([])







from setuptools import setup, find_packages

VERSION = '0.0.6'
DESCRIPTION = 'Utilities for indexing a corpus with llama-index'
LONG_DESCRIPTION = 'Utilities for indexing a corpus with llama-index'

setup(
    name='llama-kg',
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
)



from typing import Any, Awaitable, Callable, Dict, Optional, Sequence

from llama_index.core.bridge.langchain import BaseLanguageModel, BaseChatModel
from llama_index.llms.openai import OpenAI

#from llama_index.llms.llm import LLM

from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseAsyncGen,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseAsyncGen,
    CompletionResponseGen,
    LLMMetadata,
)

from llama_index.llms.openai.utils import (
    from_openai_message,
    is_chat_model,
    is_function_calling_model,
    openai_modelname_to_contextsize,
    resolve_openai_credentials,
    to_openai_message_dicts,
)

from llama_kg.llm_predictor.openai_utils import kron_openai_modelname_to_contextsize

class KronOpenAI(OpenAI):

    @property
    def metadata(self) -> LLMMetadata:
        return LLMMetadata(
            context_window=kron_openai_modelname_to_contextsize(self.model),
            num_output=self.max_tokens or -1,
            is_chat_model=is_chat_model(model=self._get_model_name()),
            is_function_calling_model=is_function_calling_model(
                model=self._get_model_name()
            ),
            model_name=self.model,
        )
    
    def _complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        #print("KronOpenAI complete called")
        response = super()._complete(prompt, **kwargs)
        text = response.text
        text = text.strip()   #triples might not start at the begining of the line
        #useful triplets are before <|endoftext|>
        text = text.split("<|endoftext|>")[0]
        #remove whitespace and all characters before first ( on all lines
        texts = text.split('\n')
        #texts = [line.strip() for line in texts]
        texts = [line[line.find('('):].strip() for line in texts]
        text = '\n'.join(texts)
        print(text)
        response.text = text
        return response



from llama_index.core import KnowledgeGraphIndex
from llama_index.core import StorageContext
from llama_index.core import ServiceContext
from llama_index.core import load_index_from_storage

from llama_index.core.graph_stores import SimpleGraphStore 
from llama_index.core.langchain_helpers.text_splitter import SentenceSplitter
#from llama_index.core.node_parser import SimpleNodeParser

from llama_kg.llm_predictor.KronOpenAILLM import KronOpenAI
from llama_kg.llm_predictor.KronLLMPredictor import KronLLMPredictor

from llama_kg.readers import S3ListReader

from pyvis.network import Network

import logging
logger = logging.getLogger(__name__)

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
s3_hook = S3Hook(aws_conn_id="minio_airflow")

def index_file(document, index, index_html_base, fs):
    logger.info(f'Indexing {document}')
    index.insert(document = document, show_progress = True)
    save_pyvis_network_graph(index, document.id_, index_html_base, fs)


def init_index(persist_path, service_context, prompt, fs):
    if not fs.exists(persist_path):
        logger.info(f'No KGIndex found for {persist_path}, creating new empty index.')
        graph_store = SimpleGraphStore()
        storage_context = StorageContext.from_defaults(graph_store=graph_store)
        index = KnowledgeGraphIndex(
            [],
            max_triplets_per_chunk=2,
            storage_context=storage_context,
            service_context=service_context,
            kg_triple_extract_template=prompt,
        )
        index.storage_context.persist(persist_dir=persist_path, fs=fs)
    else:
        print(f'Loading index from {persist_path}')
        # rebuild storage context
        storage_context = StorageContext.from_defaults(persist_dir=persist_path, fs=fs)
        # load index
        index = load_index_from_storage(storage_context=storage_context, 
                                    service_context=service_context, 
                                    max_triplets_per_chunk=2,
                                    kg_triple_extract_template=prompt,
                                    show_progress = True)
    return index


def get_service_context(model):
    # define LLM
    llm=KronOpenAI(temperature=0.01, model=model, timeout=600)
    #chunk_size+prompt_length+expected length of returned triples must be less than max_tokens
    llm.max_tokens = 512 #192-48 - 512 because of the additional instruction in the prompt
    llm_predictor = KronLLMPredictor(llm)
    print(llm_predictor.metadata)

    # define TextSplitter
    text_splitter = SentenceSplitter(chunk_size=192, chunk_overlap=48, paragraph_separator='\n')

    #define NodeParser
    #node_parser = SimpleNodeParser(text_splitter=text_splitter)
    node_parser = text_splitter

    #define ServiceContext
    service_context = ServiceContext.from_defaults(llm_predictor=llm_predictor, node_parser=node_parser)

    return service_context

def load_documents(file_list):

    loader = S3ListReader(
                    bucket='papers-kg', 
                    keys = file_list, 
                    filename_as_id = True,
                    aws_access_id=s3_hook.conn_config.aws_access_key_id, 
                    aws_access_secret=s3_hook.conn_config.aws_secret_access_key,
                    s3_endpoint_url = s3_hook.conn_config.endpoint_url,
    #                file_metadata = lambda x: {'license': license},
                )
    documents = loader.load_data()
    documents = update_documents(documents, file_list)
    return documents

def get_license(file_name, file_list):
    short_name= file_name.split('/')[-1]
    for file in file_list:
        if short_name in file:
            return file.split('/')[-2]

# update metadata and exclusions
def update_documents(documents, file_list):
    for document in documents:
        # retain the filename
        document.id_ = document.id_.split('/')[-1].split('.txt')[0]
        license = get_license(document.id_, file_list)
        document.metadata = {'license': license}
        document.excluded_embed_metadata_keys = ['license']
        document.excluded_llm_metadata_keys = ['license']
    return documents

## create pyvis graph
## use generate_html with a s3 write
def save_pyvis_network_graph(index, file_name, index_html_base, fs):
    #display all nodes
    g = index.get_networkx_graph(limit = 60000)
    net = Network(height='1000px', width='100%', notebook=True, cdn_resources="in_line", directed=True)
    net.from_nx(g)
    html_name = f'{index_html_base}/{file_name}.html'
    html = net.generate_html(html_name)
    fs.write_text(html_name, html)